--
-- SecBus ( http://secbus.telecom-paristech.fr/ ) - This file is part of SecBus
-- Copyright (C) 2009-2015 - Telecom ParisTech
-- Contacts: contact-secbus@telecom-paristech.fr
--
-- This file must be used under the terms of the CeCILL.
-- This source file is licensed as described in the file COPYING, which
-- you should have received as part of this distribution. The terms
-- are also available at
-- http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.txt
--

-- ms_cache.vhd
-- author Abdelmalek Si Merabet (abdelmalek.si-merabet@telecom-paristech.fr)
-- date : 2014-10-23
-- ms cache : set-associative  
--            write_Back policy
--            pseudo-lru
--            ram distributed 
--
-- <------------------------------------------ address------------------------- --------------------->
-- ---------------------------------------------------------------------------------------------------
-- |     Tag                                             |         set         | Block               |
-- ---------------------------------------------------------------------------------------------------
-- <--- addr_size - log2(set_size) - log2(block_size) --><-- log2(set_size) --><-- log2(block_size -->
--
-- cmd =>  100 : read 
--         010 : write dir
--         001 : write_data 

library ieee;
use ieee.std_logic_1164.all;

library global_lib;
use global_lib.numeric_std.all;
use global_lib.global.all;
use global_lib.utils.all;


entity ms_cache is
generic(
        WAYS            : natural := 2;
        SETS            : natural := 4;
        BLOCKS          : natural := 2    
);
port (
        clk             : in  std_ulogic;
        srstn           : in  std_ulogic;
        ce              : in  std_ulogic;
        
        cmd             : in  std_ulogic_vector(2 downto 0);                                           
 
        dirty_in        : in  std_ulogic;
        way_in          : in  std_ulogic_vector(WAYS - 1 downto 0);
        addr_in         : in  addr_t;                                                 
        data_in         : in  data_vector(BLOCKS     - 1 downto 0);      
        
        way_out         : out std_ulogic_vector(WAYS - 1 downto 0); 
        addr_out        : out addr_t;                                    
        data_out        : out data_vector(BLOCKS     - 1 downto 0);      
        dirty_out       : out std_ulogic;
        hit             : out std_ulogic
);
end ms_cache;


architecture rtl of ms_cache is
 
constant SET_BITS        : natural := log2_up(SETS);
constant BLOCK_BITS      : natural := log2_up(BLOCKS*CELL_SIZE);
constant TAG_BITS        : natural := ADDR_SIZE - (SET_BITS + BLOCK_BITS);   

type t_way_tag      is array (WAYS - 1 downto 0) of std_ulogic_vector(TAG_BITS - 1 downto 0);
type t_set_tag      is array (SETS - 1 downto 0) of t_way_tag;
type t_way_data     is array (WAYS - 1 downto 0) of data_vector(BLOCKS  - 1 downto 0); 
type t_set_data     is array (SETS - 1 downto 0) of t_way_data;
type t_set_dirty    is array (SETS - 1 downto 0) of std_ulogic_vector(WAYS - 1 downto 0);
type t_set_valid    is array (SETS - 1 downto 0) of std_ulogic_vector(WAYS - 1 downto 0);

type t_set_vector   is array (SETS - 1 downto 0) of  unsigned(WAYS - 1 downto 0);

signal s_tag     , r_tag   : t_set_tag;
signal s_data    , r_data  : t_set_data;
signal s_dirty   , r_dirty : t_set_vector;
signal s_valid   , r_valid : t_set_vector;
 
signal s_tag_out           : t_way_tag;                    -- data output from directory
signal s_data_out          : t_way_data;                   
signal r_lru_vector        : t_set_vector;
signal s_lru_vector        : unsigned(WAYS - 1 downto 0);  -- output from set_lru, in0 : s_lru_ways, in1 : r_lru_vector(set) 
signal s_set_dirty         : unsigned(WAYS - 1 downto 0);
signal s_hit_ways          : unsigned(WAYS - 1 downto 0);
signal s_victim_ways       : unsigned(WAYS - 1 downto 0);  -- output from mux : in0 : s_valid_victim_ways, in1 : s_lru_victim_ways, sel : s_all_valid
signal s_all_valid         : std_logic;                    -- all entries are valid for address set.
signal s_lru_ways          : unsigned(WAYS - 1 downto 0);  -- output from mux : in0 :way_in, in1 : hit_ways, sel : cmd(read,update) 
signal s_lru_enable        : std_logic;                    -- set lru when READ or UPDATE
signal s_hit               : std_logic;
signal s_not_valid_set     : unsigned(WAYS - 1 downto 0);   -- input of valid prio_1hot_encoder for a given set;
signal s_valid_set         : unsigned(WAYS - 1 downto 0);   -- input of valid prio_1hot_encoder for a given set;
signal s_not_lru_set       : unsigned(WAYS - 1 downto 0);   -- input of lru   prio_1hot_encoder for a given set;
signal s_tag_rd, r_tag_rd  : std_ulogic_vector(TAG_BITS - 1 downto 0);
signal s_cmd   , r_cmd     : std_ulogic_vector(2 downto 0);
signal r_way_in            : unsigned(WAYS - 1 downto 0);

signal s_set,   r_set      : natural range 0 to SETS - 1;

signal s_victim_addr       : addr_t;
signal s_hit_data          : data_vector(BLOCKS - 1 downto 0); 
signal s_victim_data       : data_vector(BLOCKS - 1 downto 0); 
signal s_dirty_out         : std_ulogic;

function notAllRecent(v : t_set_vector) return boolean is
begin
      for i in v'range loop
          if and_reduce(v(i)) = '1' then 
             return false;
          end if;
      end loop;

      return true;
end function;
-----------
begin  

s_tag_rd <= addr_in(addr_in'left downto (SET_BITS + BLOCK_BITS)); 
s_set    <= to_integer(unsigned(addr_in(SET_BITS + BLOCK_BITS - 1 downto BLOCK_BITS)));
s_cmd    <= cmd;
-----
genOut : for w in 0 to WAYS - 1 generate
-- Read
	s_hit_ways(w) <= r_valid(r_set)(w)  when r_tag_rd = r_tag(r_set)(w) else '0';
        s_tag_out(w)  <= r_tag(r_set)(w);
        s_data_out(w) <= r_data(r_set)(w);
end generate genOut;
------
pWay : process(s_set, cmd, r_valid, r_dirty, r_tag, r_data, way_in, addr_in, data_in, dirty_in)
begin

s_valid <= r_valid;
s_dirty <= r_dirty;
s_tag   <= r_tag;
s_data  <= r_data;

for w in 0 to WAYS - 1 loop

        if way_in(w) = '1' then        
           if cmd(1) = '1' then --write dir
              s_valid(s_set)(w) <= '1';
              s_tag(s_set)(w)   <= addr_in(addr_in'left downto (SET_BITS + BLOCK_BITS));
           end if;

           if cmd(0) = '1' then -- write data 
              s_dirty(s_set)(w) <= dirty_in;
              s_data(s_set)(w)  <= data_in;
           end if;
           exit; -- due to one hot encoding
         end if;
end loop;             
end process;

-- valid victim way
s_valid_set    <= r_valid(r_set);

-- mux lru 
s_lru_ways      <= s_hit_ways when r_cmd(2) = '1' else -- read
                   r_way_in;                             

s_lru_vector    <= set_lru(s_lru_ways, r_lru_vector(r_set));

s_not_lru_set   <= not r_lru_vector(r_set);

s_lru_enable    <= s_hit or r_cmd(1) or r_cmd(0); -- (read & hit) | write 

s_not_valid_set <= not s_valid_set;

s_all_valid     <= and_reduce(s_valid_set);    

s_hit           <= r_cmd(2) and or_reduce(s_hit_ways);                 

-- mux victim_ways
s_victim_ways <= onehot_encoder(s_not_valid_set) when s_all_valid = '0' else
                 onehot_encoder(s_not_lru_set);

-------
pclk : process(clk)
begin

if rising_edge(clk) then
    if(srstn = '0') then
        r_lru_vector   <= (others => (others => '0'));
        r_tag          <= (others => (others => (others => '0')));
        r_data         <= (others => (others => (others => (others => '0'))));
        r_valid        <= (others => (others => '0'));
        r_dirty        <= (others => (others => '0'));
        r_tag_rd       <= (others => '0');
        r_way_in       <= (others => '0');
        r_cmd          <= (others => '0');
        r_set          <= 0;

    elsif ce = '1' then

        assert notAllRecent(r_lru_vector)
        report "At least one line must be recent !!!" severity failure;

        if s_lru_enable = '1' then
            	r_lru_vector(r_set) <= s_lru_vector;
        end if;
        r_tag          <= s_tag  ;
        r_data         <= s_data ;
        r_valid        <= s_valid;
        r_dirty        <= s_dirty;
        r_tag_rd       <= s_tag_rd;  
        r_way_in       <= unsigned(way_in); 	
        r_cmd          <= s_cmd;
        r_set          <= s_set;
     end if;
end if;

end process;
-----
-- mux data out
pHitData_out : process(s_hit_ways, s_data_out)
begin

s_hit_data <= (others => (others => '0'));

for w in 0 to WAYS - 1 loop
	if s_hit_ways(w) = '1' then
                s_hit_data <= s_data_out(w);
		exit;
	end if;
end loop;

end process;

pVictim_out : process(s_victim_ways, s_tag_out, s_data_out, r_set, r_dirty)
begin

s_victim_data <= (others => (others => '0'));
s_victim_addr <= (others => '0');
s_dirty_out   <= '0';

for w in 0 to WAYS - 1 loop
	if s_victim_ways(w) = '1' then
                s_victim_addr(s_victim_addr'left downto (SET_BITS + BLOCK_BITS)) <= s_tag_out(w);
                s_victim_addr(SET_BITS + BLOCK_BITS - 1 downto BLOCK_BITS)       <= std_ulogic_vector(to_unsigned((r_set), SET_BITS));

                s_dirty_out   <= r_dirty(r_set)(w);                  
                s_victim_data <= s_data_out(w);
		exit;
	end if;
end loop;

end process;

hit       <= s_hit;
way_out   <= std_ulogic_vector(s_hit_ways)    when s_hit = '1' else
             std_ulogic_vector(s_victim_ways);           

addr_out  <= s_victim_addr;
dirty_out <= s_dirty_out;
data_out  <= s_hit_data    when s_hit = '1' else
             s_victim_data;
end rtl;


