--
-- SecBus ( http://secbus.telecom-paristech.fr/ ) - This file is part of SecBus
-- Copyright (C) 2009-2015 - Telecom ParisTech
-- Contacts: contact-secbus@telecom-paristech.fr
--
-- This file must be used under the terms of the CeCILL.
-- This source file is licensed as described in the file COPYING, which
-- you should have received as part of this distribution. The terms
-- are also available at
-- http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.txt
--

-- mt_cache.vhd
-- author Abdelmalek Si Merabet (abdelmalek.si-merabet@telecom-paristech.fr)
-- date : 2014-10-23
-- ms cache : set-associative  
--            write_Back policy
--            pseudo-lru
--            ram distributed 
--
-- <------------------------------------------ address------------------------- --------------------->
-- ---------------------------------------------------------------------------------------------------
-- |     Tag                                             |         set         | Block               |
-- ---------------------------------------------------------------------------------------------------
-- <--- addr_size - log2(set_size) - log2(block_size) --><-- log2(set_size) --><-- log2(block_size -->
--
-- cmd => 100 : inval 
--        010 : read 
--        001 : write 

library ieee;
use ieee.std_logic_1164.all;

library global_lib;
use global_lib.numeric_std.all;
use global_lib.global.all;
use global_lib.utils.all;


entity mt_cache is
generic(
        WAYS            : natural := 4;
        SETS            : natural := 16; 
        BLOCKS          : natural := 2;    
        TSH             : natural := 2  -- associativity/2
);
port (
        clk             : in  std_ulogic;
        srstn           : in  std_ulogic;
        ce              : in  std_ulogic;
        
        cmd             : in  std_ulogic_vector(2 downto 0); -- 000 : nop; 100 : inval; 010 : read; 001 : write         
 
        dirty_in        : in  std_ulogic;
        wdir_in         : in  std_ulogic;
        wdata_in        : in  std_ulogic;
        updTsh_in       : in std_ulogic;
        incTsh_in       : in std_ulogic;
        way_in          : in  std_ulogic_vector(WAYS - 1 downto 0);
        addr_in         : in  addr_t;                                                 
        data_in         : in  data_vector(BLOCKS     - 1 downto 0);      
        
        way_out         : out std_ulogic_vector(WAYS - 1 downto 0); 
        addr_out        : out addr_t;                                    
        data_out        : out data_vector(BLOCKS     - 1 downto 0);      
        dirty_out       : out std_ulogic;
        hit             : out std_ulogic;
        full            : out std_ulogic
);
end mt_cache;


architecture rtl of mt_cache is
 
constant SET_BITS        : natural := log2_up(SETS);
constant BLOCK_BITS      : natural := log2_up(BLOCKS*CELL_SIZE);
constant TAG_BITS        : natural := ADDR_SIZE - (SET_BITS + BLOCK_BITS);   

type t_way_tag      is array (WAYS - 1 downto 0) of std_ulogic_vector(TAG_BITS - 1 downto 0);
type t_set_tag      is array (SETS - 1 downto 0) of t_way_tag;
type t_way_data     is array (WAYS - 1 downto 0) of data_vector(BLOCKS  - 1 downto 0); 
type t_set_data     is array (SETS - 1 downto 0) of t_way_data;
type t_set_dirty    is array (SETS - 1 downto 0) of std_ulogic_vector(WAYS - 1 downto 0);
type t_set_valid    is array (SETS - 1 downto 0) of std_ulogic_vector(WAYS - 1 downto 0);
type t_tsh          is array (SETS - 1 downto 0) of natural range 0 to TSH;
type t_set_vector   is array (SETS - 1 downto 0) of  unsigned(WAYS - 1 downto 0);

signal s_tag     , r_tag   : t_set_tag;
signal s_data    , r_data  : t_set_data;
signal s_dirty   , r_dirty : t_set_vector;
signal s_valid   , r_valid : t_set_vector;
 
signal s_set_dirty         : unsigned(WAYS - 1 downto 0);
signal s_hit_ways          : unsigned(WAYS - 1 downto 0);
signal s_victim_ways       : unsigned(WAYS - 1 downto 0);  -- output from mux : in0 : s_valid_victim_ways, in1 : s_dirty_victim_ways, sel : s_all_valid
signal s_all_valid         : std_logic;                    -- all entries are valid for address set.
signal s_hit               : std_logic;
signal s_not_valid_set     : unsigned(WAYS - 1 downto 0);   
signal s_not_dirty_set     : unsigned(WAYS - 1 downto 0);   
signal s_valid_set         : unsigned(WAYS - 1 downto 0);   
signal s_dirty_set         : unsigned(WAYS - 1 downto 0);   
signal s_tag_rd, r_tag_rd  : std_ulogic_vector(TAG_BITS - 1 downto 0);
signal s_cmd   , r_cmd     : std_ulogic_vector(2 downto 0);
signal r_way_in            : unsigned(WAYS - 1 downto 0);

signal s_set,   r_set      : natural range 0 to SETS - 1;

signal s_victim_addr       : addr_t;
signal s_hit_data          : data_vector(BLOCKS - 1 downto 0); 
signal s_dirty_out         : std_ulogic;
signal s_full              : std_ulogic;

signal s_tsh,  r_tsh       : t_tsh;

function notAllRecent(v : t_set_vector) return boolean is
begin
      for i in v'range loop
          if and_reduce(v(i)) = '1' then 
             return false;
          end if;
      end loop;

      return true;
end function;
-----------
begin  

s_tag_rd <= addr_in(addr_in'left downto (SET_BITS + BLOCK_BITS)); 
s_set    <= to_integer(unsigned(addr_in(SET_BITS + BLOCK_BITS - 1 downto BLOCK_BITS)));
s_cmd    <= cmd;
-----
genOut : for w in 0 to WAYS - 1 generate
-- Read
      s_hit_ways(w) <= r_valid(r_set)(w)  when r_tag_rd = r_tag(r_set)(w) else '0';
end generate genOut;
------
pWrite : process(s_set, cmd, 
               r_valid, r_dirty, r_tag, r_data, way_in, addr_in, data_in, dirty_in,
               wdir_in, wdata_in)
begin

s_valid <= r_valid;
s_dirty <= r_dirty;
s_tag   <= r_tag;
s_data  <= r_data;

for w in 0 to WAYS - 1 loop

        if way_in(w) = '1' then
              if cmd(2) = '1' then      -- inval
                   s_valid(s_set)(w) <= '0';
                   s_dirty(s_set)(w) <= '0';

              elsif cmd(0) = '1' then   -- write 
                   if wdir_in = '1' then  
                      s_tag(s_set)(w)   <= addr_in(addr_in'left downto (SET_BITS + BLOCK_BITS));
                      s_valid(s_set)(w) <= '1';
                   end if;
                   
                   if wdata_in = '1' then
                      s_data(s_set)(w)  <= data_in;
                      s_dirty(s_set)(w) <= dirty_in;
                   end if;
                   
              end if;
              exit; -- one hot encoding
         end if;
end loop;             
end process;

-- valid victim way
s_valid_set    <= r_valid(r_set);
s_dirty_set    <= r_dirty(r_set);


s_not_valid_set <= not s_valid_set;
s_not_dirty_set <= not s_dirty_set;

s_all_valid     <= and_reduce(s_valid_set);    

s_hit           <= r_cmd(1) and or_reduce(s_hit_ways);                 

-- mux victim_ways
s_victim_ways <= onehot_encoder(s_not_valid_set) when s_all_valid = '0' else
                 onehot_encoder(s_not_dirty_set);


pUpdTsh : process(updTsh_in, incTsh_in, r_set, r_tsh)
begin

s_tsh <= r_tsh;

if updTsh_in = '1' then
   if incTsh_in = '1' then
     s_tsh(r_set) <= r_tsh(r_set) + 1;
   else
     s_tsh(r_set) <= r_tsh(r_set) - 1;
   end if;
end if;
end process;

-------
pclk : process(clk)
begin

if rising_edge(clk) then
    if(srstn = '0') then
        r_tag          <= (others => (others => (others => '0')));
        r_data         <= (others => (others => (others => (others => '0'))));
        r_valid        <= (others => (others => '0'));
        r_dirty        <= (others => (others => '0'));
	r_tag_rd       <= (others => '0');
	r_way_in       <= (others => '0');
	r_cmd          <= (others => '0');
	r_set          <= 0;
	r_tsh          <= (others => 0);

     elsif ce = '1' then
        r_tag          <= s_tag  ;
        r_data         <= s_data ;
        r_valid        <= s_valid;
        r_dirty        <= s_dirty;
        r_tag_rd       <= s_tag_rd;  
	r_way_in       <= unsigned(way_in); 	
	r_cmd          <= s_cmd;
	r_set          <= s_set;
	r_tsh          <= s_tsh;
     end if;
end if;

end process;
-----
-- mux data out
pHitData_out : process(s_hit_ways, r_set, r_data, r_dirty)
begin

s_hit_data <= (others => (others => '0'));

for w in 0 to WAYS - 1 loop
	if s_hit_ways(w) = '1' then
                s_hit_data  <= r_data(r_set)(w);
                s_dirty_out <= r_dirty(r_set)(w);                  
		exit;
	end if;
end loop;

end process;


pVictim_out : process(r_tag, r_set, s_hit_ways, r_dirty)
begin

s_victim_addr <= (others => '0');
-- if threshold reached => victim address = first address of a dirty block but the one we're writing.
for w in 0 to WAYS - 1 loop
	if r_dirty(r_set)(w) = '1' and s_hit_ways(w) = '0' then
                s_victim_addr(s_victim_addr'left downto (SET_BITS + BLOCK_BITS)) <= r_tag(r_set)(w);
                s_victim_addr(SET_BITS + BLOCK_BITS - 1 downto BLOCK_BITS)       <= std_ulogic_vector(to_unsigned((r_set), SET_BITS));
		exit;
	end if;
end loop;

end process;

pFull : process(r_set, r_tsh)
begin

s_full <= '0';

if r_tsh(r_set) = TSH then -- or >=
   s_full <= '1';
end if;

end process;

hit       <= s_hit;
way_out   <= std_ulogic_vector(s_hit_ways)    when s_hit = '1' else
             std_ulogic_vector(s_victim_ways);           

addr_out  <= s_victim_addr;
dirty_out <= s_dirty_out;
data_out  <= s_hit_data;
full      <= s_full;

end rtl;


